/** @file    Automated tests for the the Blowfish implementation.

   @author   Ivan Vecerina - 2002 - This source file is hereby placed in the public domain

Some test cases are derived from an implementation of Blowfish
 by George Anescu, found at <http://www.codeproject.com/cpp/blowfish.asp>.
*/
#ifndef NDEBUG // tests not included in release build

#include "blowfish.h"
#include <iostream>


namespace {


/// Outputs the hexadeximal representation of @a byte into @a dst.
void byteToHex(const uint8 byte, char dstHex[2])
{
   static const char xdigits[] = "0123456789ABCDEF";
   dstHex[0] = xdigits[ (byte/16) & 15 ];
   dstHex[1] = xdigits[  byte     & 15 ];
}

/**
@pre @c must be a valid hexadecimal character.
Assumptions made:
 - ('0' & 0xF) == 0
 - '0' to '9' have consecutive char codes
 - 'A' to 'F' have consecutive char codes
And for lowercase hexadecimal values to work:
 - 'a' to 'f' have consecutive char codes, and ('A'&0xF) == ('a'&0xF)
*/
inline int valueOfHexDigit(char c)
{
   enum { alphaOffs = 10 - ('A'&0xF) };
   if( c > '9' ) c += alphaOffs;
   return (c & 0xF);
}

/// Converts two hexadecimal digits into an unsigned value.
void hexToByte(const char src[2], uint8& dstByte)
{
   dstByte = ( valueOfHexDigit(src[0]) << 4 ) + valueOfHexDigit(src[1]);
}

/// Prints the memory range defined by @a data and @a size to @a dst, in hexadecimal form.
void printBufAsHex(std::ostream& dst, uint8 const* data, unsigned size)
{
   while( size-- )
   {
      char hh[2];
      byteToHex( *data++, hh );
      dst << hh[0] << hh[1];
   }
}

/// Adapter to send a data buffer to a stream in hexadecimal form.
class BufAsHex
{
   uint8 const*  const data_;
   unsigned long const size_;
public:
   BufAsHex( void const* data, unsigned long size )
    : data_( reinterpret_cast<uint8 const*>(data) ), size_(size) {}

   friend std::ostream& operator << ( std::ostream& dst, BufAsHex const& buf )
   { printBufAsHex( dst, buf.data_, buf.size_ ); return dst; }
};


/// Loads the contents of a data buffer from a source encoded as a hexadecimal string.
void loadBufferFromHex( uint8* pucCharStr, unsigned size, char const* pszHexStr )
{
   for( unsigned i = 0 ; i != size ; ++i )
		hexToByte( &pszHexStr[2*i], pucCharStr[i] );
}


/** Definition of a test vector ( key, source, and outputs in various encryption modes ).
All data vectors are stored as hexadecimal strings.
*/
struct TestData {
   const char* key; ///< Encryption key (or password).
   const char* src; ///< Source data to be encrypted.
   const char* ecb; ///< Data encrypted with Blowfish in ECB mode.
   const char* cbc; ///< Data encrypted with Blowfish in CBC mode.
   const char* cfb; ///< Data encrypted with Blowfish in CFB mode.
};


/** Test data used to check the consistency of the current algorithms.

@NOTE: The output of this program is test data with completed results if some were incomplete
       => program output can be copy-pasted below to extend the tests
*/
TestData testData[] = {
   // the original test data provided in Paul Kocher's implementation
   {"544553544B4559"//"TESTKEY"
      , "0000000100000002" // L=1, R=2
      , "DF333FD230A71BB4" // original test result.
      , "DF333FD230A71BB4", "11D337059796D47A" // added results for other modes
     },
//    Initial Key          Source Data      ECB-Encrypted Data  CBC-Encrypted Data  CFB-Encrypted Data
   {"0000000000000000", "0000000000000000", "4EF997456198DD78", "4EF997456198DD78", "4EF997456198DD78"},
   {"FFFFFFFFFFFFFFFF", "FFFFFFFFFFFFFFFF", "51866FD5B85ECB8A", "51866FD5B85ECB8A", "0DE1658848E3B643"},
   {"3000000000000000", "1000000000000001", "7D856F9A613063F2", "7D856F9A613063F2", "E4A584A4E46E60C4"},
   {"1111111111111111", "1111111111111111", "2466DD878B963C9D", "2466DD878B963C9D", "C30C546E25ED1F11"},
   {"0123456789ABCDEF", "1111111111111111", "61F9C3802281B096", "61F9C3802281B096", "354857994645278B"},
   {"1111111111111111", "0123456789ABCDEF", "7D0CC630AFDA1EC7", "7D0CC630AFDA1EC7", "D33E0018BD57C3EF"},
   {"0000000000000000", "0000000000000000", "4EF997456198DD78", "4EF997456198DD78", "4EF997456198DD78"},
   {"FEDCBA9876543210", "0123456789ABCDEF", "0ACEAB0FC6A0A28D", "0ACEAB0FC6A0A28D", "2207DD0A2FDDBC73"},
   {"7CA110454A1A6E57", "01A1D6D039776742", "59C68245EB05282B", "59C68245EB05282B", "F3FC9D0AF9DBC726"},
   {"0131D9619DC1376E", "5CD54CA83DEF57DA", "B1B8CC0B250F09A0", "B1B8CC0B250F09A0", "D9E00763BA611EE8"},
   {"07A1133E4A0B2686", "0248D43806F67172", "1730E5778BEA1DA4", "1730E5778BEA1DA4", "72D6B43FCD314266"},
   {"3849674C2602319E", "51454B582DDF440A", "A25E7856CF2651EB", "A25E7856CF2651EB", "275E44A81A5B9D6C"},
   {"04B915BA43FEB5B6", "42FD443059577FA2", "353882B109CE8F1A", "353882B109CE8F1A", "AF687350FD46D051"},
   {"0113B970FD34F2CE", "059B5E0851CF143A", "48F4D0884C379918", "48F4D0884C379918", "5F9B08ED9781356C"},
   {"0170F175468FB5E6", "0756D8E0774761D2", "432193B78951FC98", "432193B78951FC98", "95C8903E3DBEEF21"},
   {"43297FAD38E373FE", "762514B829BF486A", "13F04154D69D1AE5", "13F04154D69D1AE5", "1398FD0002A34FC3"},
   {"07A7137045DA2A16", "3BDD119049372802", "2EEDDA93FFD39C79", "2EEDDA93FFD39C79", "082BA912754694BF"},
   {"04689104C2FD3B2F", "26955F6835AF609A", "D887E0393C2DA6E3", "D887E0393C2DA6E3", "4DB176CC15215208"},
   {"37D06BB516CB7546", "164D5E404F275232", "5F99D04F5B163969", "5F99D04F5B163969", "3BB2C9DC72FD7FC9"},
   {"1F08260D1AC2465E", "6B056E18759F5CCA", "4A057A3B24D3977B", "4A057A3B24D3977B", "8E18AB1C3B281422"},
   {"584023641ABA6176", "004BD6EF09176062", "452031C1E4FADA8E", "452031C1E4FADA8E", "783DE3902A274486"},
   {"025816164629B007", "480D39006EE762F2", "7555AE39F59B87BD", "7555AE39F59B87BD", "C85842046B074372"},
   {"49793EBC79B3258F", "437540C8698F3CFA", "53C55F9CB49FC019", "53C55F9CB49FC019", "03967D6BC764761F"},
   {"4FB05E1515AB73A7", "072D43A077075292", "7A8E7BFA937E89A3", "7A8E7BFA937E89A3", "33D0061D1E09F942"},
   {"49E95D6D4CA229BF", "02FE55778117F12A", "CF9C5D7A4986ADB5", "CF9C5D7A4986ADB5", "5FEA975EFBB30BFB"},
   {"018310DC409B26D6", "1D9D5C5018F728C2", "D1ABB290658BC778", "D1ABB290658BC778", "297ED0086339DE3F"},
   {"1C587F1C13924FEF", "305532286D6F295A", "55CB3774D13EF201", "55CB3774D13EF201", "B8E1A27FFC6DAAD9"},
   {"0101010101010101", "0123456789ABCDEF", "FA34EC4847B268B2", "FA34EC4847B268B2", "60158372A70C084B"},
   {"1F1F1F1F0E0E0E0E", "0123456789ABCDEF", "A790795108EA3CAE", "A790795108EA3CAE", "E040F3F09DF3371F"},
   {"E0FEE0FEF1FEF1FE", "0123456789ABCDEF", "C39E072D9FAC631D", "C39E072D9FAC631D", "B077B163CAEF2F97"},
   {"0000000000000000", "FFFFFFFFFFFFFFFF", "014933E0CDAFF6E4", "014933E0CDAFF6E4", "B10668BA9E672287"},
   {"FFFFFFFFFFFFFFFF", "0000000000000000", "F21E9A77B71C49BC", "F21E9A77B71C49BC", "F21E9A77B71C49BC"},
   {"0123456789ABCDEF", "0000000000000000", "245946885754369A", "245946885754369A", "245946885754369A"},
   {"FEDCBA9876543210", "FFFFFFFFFFFFFFFF", "6B5C5A9C5D9E0A5A", "6B5C5A9C5D9E0A5A", "DCDB679259898E63"},
   // longer data sequences added to test mode-dependent variations of encrypted output
   {"1234", "6B5C5A9C5D9E0A5A6B5C5A9C5D9E0A5ADCDB679259898E63"
          , "C46225685EC8F311C46225685EC8F311AAE49F0647641B94"
          , "C46225685EC8F311D7D0ADCD282F13F155A16118C9C6B6F4"
          , "07C202628A7D94775D6FA8AC6E667DDB9EADF5D9AFC9E8A9"},
   {"567890", "6B5C5A9C5D9E0A5A6B5C5A9C5D9E0A5ADCDB679259898E634A057A3B24D3977B"
            , "AF509E758F314B1CAF509E758F314B1C9A4B6D6D5A0807739AD15FD746593B2D"
            , "AF509E758F314B1C3F133F40F8F55374FD4A7FFF0A8809BA6BC4698C61D5FC88"
            , "171265AA5AB49793DAB717B8698F4378BCD809ADCDECFECBB588DA1AE2C1A921"},
   {00} // end mark
};


/// List of encryption modes.
enum EMode
      { ECB ///< Direct encryption of independent blocks (no data chaining).
      , CBC ///< Encrypt after xor-ing source with previous encrypted result.
      , CFB ///< Destination is source xor-ed with encrypted previous result.
      };

/// Returns the encription result from @a data corresponding to encryption @a mode.
const char* getResultForMode(TestData const& data, EMode mode)
{
   switch( mode ) {
      case ECB: return data.ecb;
      case CBC: return data.cbc;
      case CFB: return data.cfb;
      default: throw(-1);
   }
}

void encryptInMode( void const* src, void* dst, uint32 len, blowfish::Pad& pad, EMode mode )
{
   using namespace blowfish;
   switch( mode ) {
      case ECB:  encrypt_ECB( pad, src, dst, len );  return;
      case CBC:  encrypt_CBC( pad, src, dst, len );  return;
      case CFB:  encrypt_CFB( pad, src, dst, len );  return;
      default: throw(-1);
   }
}

void decryptInMode( void const* src, void* dst, uint32 len, blowfish::Pad& pad, EMode mode )
{
   using namespace blowfish;
   switch( mode ) {
      case ECB:  decrypt_ECB( pad, src, dst, len );  return;
      case CBC:  decrypt_CBC( pad, src, dst, len );  return;
      case CFB:  decrypt_CFB( pad, src, dst, len );  return;
      default: throw(-1);
   }
}


}//namespace


void test_blowfish()
{
   bool resultsIncomplete = false;
   enum { maxDataSize = 256 };

	for( int dataInd=0 ; !! testData[dataInd].key ; ++dataInd )
	{
      unsigned char keyBuf[maxDataSize];
      unsigned      keyLen;
      {
         char const* key = testData[dataInd].key;
         keyLen = strlen( key )/2;
		   loadBufferFromHex( keyBuf, keyLen, key );
         std::cout << "   {\"" << key << '\"';
      }

      unsigned char sourceDataBuf[maxDataSize];
      unsigned      dataLen;
      {
         char const* src = testData[dataInd].src;
         dataLen = strlen( src )/2;
		   loadBufferFromHex( sourceDataBuf, dataLen, src );
         std::cout << ", \"" << src << '\"';
      }

      for( EMode mode = ECB ; mode <= CFB ; mode = EMode(mode+1) )
      {
         unsigned char encrypted[maxDataSize];
         {
		      blowfish::Pad pad = blowfish::generatePad(keyBuf, keyLen);
		      encryptInMode( sourceDataBuf, encrypted, dataLen, pad, mode );
            std::cout << ", \"" << BufAsHex(encrypted,dataLen) << '\"';
         }

         unsigned char restored[maxDataSize];
         {
		      blowfish::Pad pad = blowfish::generatePad(keyBuf, keyLen);
		      decryptInMode( encrypted, restored, dataLen, pad, mode );
         }

         for( unsigned i = 0 ; i != dataLen ; ++i )
         {
            if( sourceDataBuf[i] != restored[i] )
            {
               std::cout << "encryption-decryption cycle is broken" << std::endl;
               throw 3;   // encrypt-decrypt cycle is broken
            }
         }


         unsigned char origEncrypted[maxDataSize];
         {
		      const char* hex = getResultForMode(testData[dataInd], mode);
            if( ! hex ) { resultsIncomplete = true; continue; }
            if( strlen(hex)/2 != dataLen ) exit(-1);
		      loadBufferFromHex( origEncrypted, dataLen, hex );
         }

         for( unsigned j = 0 ; j != dataLen ; ++j )
         {
            if( encrypted[j] != origEncrypted[j] )
            {
               std::cout << "Encrypted data does not match previous results." << std::endl;
               throw 2;   // does not match previous results
            }
         }
      }

      std::cout << "}," << std::endl;
	}

   if( resultsIncomplete )
   {
      std::cout << "\n\n   ------- WARNING --------\n"
                   "All went well, but results were incomplete in the test data.\n"
                   "Please copy-paste the ouput above into the source code test data,\n"
                   "to allow extended regression testing in the future.\n\n" << std::endl;
      throw 1;
   }
}



/*
int main()
{
   try {
      test_blowfish();
      return 0;
   }
   catch(int code)   { return code; }
   catch(...)        { std::cout<<"Unexpected error"<<std::endl; return -1; }
}
*/


#endif //NDEBUG
